/*
 *  Canon Inkjet Printer Driver for Linux
 *  Copyright CANON INC. 2009
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// Linux System Header
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/stat.h>

// CUPS Header
#include <cups/cups.h>
#include <cups/ipp.h>

/* lgmon command length		 */
#define LGMON_DATA_WRITE_STR_LEN		8192	/* device write data length	 */
#define	LGMON_CMD_BASE_STR				PROG_PATH "/cnijnetprn"

int     g_pid = -1;
int     g_signal = 0;

/* Messagess table */
enum {
	CANON_STS_SUCCESS,
	CANON_STS_NG	 ,
	CANON_STS_PIPE_ERR,
	CANON_STS_FORK_ERR,
	CANON_STS_LGMON_EXEC_ERR,
	CANON_STS_PRINT_CANCEL,
	CANON_STS_DEFAULT
} ;

// Header file for CANON
#include "cnijnetlm.h"

/*--------- Prototypes.	*/
static void lgmon_sigterm_handler( int ) ;
static int  net_lgmon_exec( int, int, FILE *, int , char *  , char *  , char * ) ;
static void data_write( int, FILE *, int , int ) ;

/*----------------------------------------------------------------------------*
 * This function is a main function for calling a monitor process. 
 */

int 	_net_canon_bj( int argc, int mode, FILE *fp, int copies, char* option, char *resource, char *user, char *document)
{
#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
    struct  sigaction action;			/* Actions for POSIX signals        */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

	int sts ;									/* function status		*/
	int utilbyte = 0;

/*
* When a signal (SIGTERM/SIGPIPE) is received, printing process
* and the lgmon process will be terminated.
*/


#ifdef HAVE_SIGSET
	sigset(SIGPIPE, SIG_IGN);
	sigset(SIGTERM, lgmon_sigterm_handler);
#elif defined(HAVE_SIGACTION)
	memset(&action, 0, sizeof(action));
	action.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &action, NULL);

	sigemptyset(&action.sa_mask);
	sigaddset(&action.sa_mask, SIGTERM);
	
	action.sa_handler = lgmon_sigterm_handler;
	sigaction(SIGTERM, &action, NULL);
#else
	signal(SIGPIPE, SIG_IGN);
	signal(SIGTERM, lgmon_sigterm_handler);
#endif

	/* Clear printer status */
	setbuf(stderr, NULL);
	//fprintf( stderr, "INFO: \n" );
	
	sts = net_lgmon_exec( argc, mode, fp, copies, resource, user, document);
	
	/* Clear printer status */
	//fprintf( stderr, "INFO: \n" );
	
	return( sts ) ;
}


/*----------------------------------------------------------------------------*
 * When a signal (SIGTERM) is received, printing process
 * and the lgmon process will be terminated.
 */
static
void lgmon_sigterm_handler(int sig_code)
{
	/* Clear printer status */
	fprintf( stderr, "INFO: \n" );

	if( g_pid > 0 )
		kill(g_pid, SIGTERM);
	
	g_signal = 1;
}


/*----------------------------------------------------------------------------*
 * This function is carrying out processing which calls a lgmon process. 
 * And pipes create.
 *
 * Attention: param(char *command) is unused.
 */
static
int		net_lgmon_exec( int argc, int mode, FILE *fp, int copies , char *resource, char *user, char *document ) {
	int		rtn ;								/* return status buffer		*/
	int		pipe_fds[2] ;						/* pipe disripter			*/
	struct  sigaction action;

	if	( pipe(pipe_fds) < 0 ) {				/* pipe cleate				*/
		return( CANON_STS_PIPE_ERR ) ;
	}

	g_pid = fork() ;							/* fork()					*/

	if	( g_pid == -1 ) {    		          	/* fork() error 			*/
		return( CANON_STS_FORK_ERR ) ;
	}
	else if ( g_pid == 0 ) {					/* child process 			*/
		static	int		cmd_argv_cnt = 0 ;		/* argument counter			*/
		static	char	*cmd_argv[16] ;			/* command argument beffer	*/
		static  char	*cmd_sh_path = "/bin/sh";	/* shell path			*/
		static	char	str_msg[128] ;			/* message beffer			*/
		memset( str_msg, 0x00, sizeof(str_msg) ) ;
		
		if (mode == CNNET_MODE_SEARCH){
			close(0);
			dup2(pipe_fds[0], 0);
			close(pipe_fds[0]);
			close(pipe_fds[1]);
			
			cmd_argv[cmd_argv_cnt++] = NULL ;
		} else {
			close(0);
			close(1);
			dup2(pipe_fds[0], 0);
			close(pipe_fds[0]);
			close(pipe_fds[1]);
			
			cmd_argv[cmd_argv_cnt++] = resource ;		/* create lgmon command 	*/
			cmd_argv[cmd_argv_cnt++] = user ;
			cmd_argv[cmd_argv_cnt++] = document ;
			cmd_argv[cmd_argv_cnt++] = NULL ;
		}
		rtn = execvp ( LGMON_CMD_BASE_STR, cmd_argv ) ;/* lgmon exec 				*/

		if  ( rtn == -1 ) {						/* error					*/
			switch( errno ) {
				case    ENOENT :				/* Command not found 		*/
					fprintf( stderr, "ERROR: Command not found \n") ;
					break ;
				case    EACCES :				/* Command Permission denied*/
					fprintf( stderr, "ERROR: Command Permission denied \n") ;
					break ;
				default :
					fprintf( stderr, "ERROR: Please refer to <asm/errno.h>\n") ;
			}
			return( CANON_STS_NG ) ;
		}
	}
	else {										/* parents process			*/
		close( pipe_fds[0] ) ;					/* pipe(in) close           */
		if (mode != CNNET_MODE_SEARCH){
			data_write( argc, fp, copies, pipe_fds[1] ) ;	/* data write 	*/
		}
		close( pipe_fds[1] ) ;					/* pipe(out) close          */
	}
	
	
	if	( g_pid > 0 ) {				/* if child process has started....	*/
		int child_status = 0;
		pid_t child_pid = 0;
		
		// waitpid( g_pid, NULL, 0) ;		/* child process wait		*/
		child_pid = wait(&child_status);
		
		if (!WIFEXITED(child_status)){
			return( CANON_STS_NG ) ;
		}
		rtn = WEXITSTATUS(child_status);
		
		if (rtn == 0){
			return( CANON_STS_SUCCESS ) ;
		} else {
			return( CANON_STS_NG ) ;
		}
	}
	
	return( CANON_STS_SUCCESS ) ;
}

/*----------------------------------------------------------------------------*
 *  This function writes data in a device (printer).
 */
static
void	data_write( int argc, FILE *fp, int copies, int pipe_fds )
{

	int			wbytes;     				/* Number of bytes written		*/
	size_t		nbytes,     				/* Number of bytes read			*/
				tbytes;						/* Total number of bytes written*/
	char		buffer[LGMON_DATA_WRITE_STR_LEN],		/* Output buffer	*/
				*bufptr;    				/* Pointer into buffer			*/

	while (copies > 0) {

		copies --;

		if  ( fp != stdin ) {
			rewind(fp);
		}

		tbytes = 0;

		/*
		*	data read
		*/
		while ((nbytes = fread(buffer, 1, sizeof(buffer), fp)) > 0) {

			if( g_signal != 0 )			/* signal(SIGTERM) writing end. */
				break;

			/*
			* Write the print data to the printer...
			*/
			tbytes += nbytes;
			bufptr = buffer;

			while (nbytes > 0) {
				if ((wbytes = write(pipe_fds, bufptr, nbytes)) < 0)
					if (errno == ENOTTY)			/* retry. 				*/
						wbytes = write(pipe_fds, bufptr, nbytes);

				if (wbytes < 0) {					/* write error			*/
					perror("ERROR: Unable to send print file to printer");
					break;
				}

				nbytes -= wbytes;
				bufptr += wbytes;

				if( g_signal != 0 )		/* signal(SIGTERM) writing end. */
					break;
			}
		}

		if( g_signal != 0 )				/* signal(SIGTERM) writing end. */
			break;

	} /* while end */

	if( g_signal != 0 ) {
		// read all
		while ((nbytes = fread(buffer, 1, sizeof(buffer), fp)) > 0);
		// flush
		fflush(fp);
	}
}

