/*
 *  Canon Inkjet Printer Driver for Linux
 *  Copyright CANON INC. 2009
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>

#include <libcnnet.h>

#define  CACHE_PATH     BJLIB_PATH "/cnnet.ini"

/*******************************************************************************
	Definition of time
********************************************************************************/

#define RETRYCOUNT_FIND_PRINTERS (5)
#define WAITTIME_FIND_PRINTERS   (30)
#define WAITTIME_START_SESSION   (5)
#define WAITTIME_RETRY_SEARCH    (30)

#define MAX_COUNT_OF_PRINTERS   (64)
#define STRING_SHORT            (32)

//#define RETRY_COUNT             (5)
//#define WAITTIME_AT_RETRY       (1000)

#define SESSION_TIMEOUT         (30)
#define PACKET_BUFFER_SIZE      (4)

/*******************************************************************************
	Definition of mode
********************************************************************************/

#define CNNL_JOB_PREPARING      (0x0001)
#define CNNL_JOB_WAITING        (0x0002)
#define CNNL_JOB_PRINTING       (0x0003)
#define CNNL_JOB_FINISHING      (0x0004)

#define CNNL_JOB_CANCELLED      (0x1000)
#define CNNL_JOB_ERROR          (0x2000)

/*******************************************************************************
	Definitions of variable
********************************************************************************/

CNNLHANDLE    hnd  = NULL;
volatile int  mode = CNNL_JOB_PREPARING;


/*******************************************************************************
	Definitions of function
********************************************************************************/

static uint32_t fread_with_retry(char *rbuffer, unsigned long bufsize, FILE *fp);
static int  _ij_network_print(const char *, const char *, const char *, FILE *);
static void _ij_network_terminate(int);
static int  _ij_network_search(void);

/*******************************************************************************
	Main
********************************************************************************/
int main(int argc, char **argv){
	int retcode = CNNL_RET_SUCCESS;
	struct sigaction action_pipe, action_term;
	
	/* ignore SIGPIPE */
	memset(&action_pipe, 0, sizeof(action_pipe));
	action_pipe.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &action_pipe, NULL);
	
	/* set signal handler for SIGTERM */
	memset(&action_term, 0, sizeof(action_term));
	sigemptyset(&action_term.sa_mask);
	sigaddset(&action_term.sa_mask, SIGTERM);
	action_term.sa_handler = _ij_network_terminate;
	action_term.sa_flags = 0;
	sigaction(SIGTERM, &action_term, NULL);

	if (argc > 1){
		retcode = _ij_network_print(argv[0], argv[1], argv[2], stdin);
	} else {
		_ij_network_search();
	}

	switch (retcode){
		case 0:
			exit(0);
		default:
			exit(1);
	}
}

/*******************************************************************************
	fread_with_timeout
********************************************************************************/
static uint32_t fread_with_retry(char *rbuffer, unsigned long bufsize, FILE *fp){
	uint32_t d_time;
	fd_set  fds;
	struct timeval timeout;
	int errorcount=0;
	
	while (!(mode & CNNL_JOB_CANCELLED)){
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		
		FD_ZERO(&fds);
		FD_SET(0,&fds);
		select(1,&fds,NULL,NULL,&timeout);
		
		if(FD_ISSET(0,&fds)){
			size_t s = fread(rbuffer, sizeof(char), bufsize, fp);
			
			if (ferror(fp) == 0){
				return (s * sizeof(char));
			} else {
				mode |= CNNL_JOB_ERROR;
				return 0;
			}
		} else {
			// keep connection
			if (CNNL_GetTimeout(hnd, &d_time, 3, 3000) != CNNL_RET_SUCCESS){
				mode |= CNNL_JOB_ERROR;
				return 0;
			} else {
				errorcount=0;
			}
		}
	}
	return 0;
}

/*******************************************************************************
	_ij_network_print
********************************************************************************/
static int _ij_network_print(
	const char  *macaddr,
	const char  *user,
	const char  *job,
	FILE *fp
){
	int errcode = 0, errcount=0; // , tmp_ip[4];
	char ipaddr[STRING_SHORT], hostname[STRING_SHORT];
	
	int ret;
	unsigned long written_size=0, bufsize=1024*PACKET_BUFFER_SIZE;
	char *rbuffer, *p=NULL;
	uint32_t len=0, wlen=0;
	
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);
	memset(ipaddr, 0x00, STRING_SHORT);
	memset(hostname, 0x00, STRING_SHORT);
	
	// allcate memory
	if ((rbuffer=(char*)malloc(sizeof(char)*bufsize)) == NULL){
		goto error_init;
	}
	
	// get hostname
	if (gethostname(hostname, STRING_SHORT-1) != 0){
		errcode = 1;
		fprintf(stderr, "ERROR: cannot get hostname\n");
		goto error_init;
	}
	mode = CNNL_JOB_PREPARING;
	
	/* write to printer */
	{
find_printer:
		/* ---------------------------------------------------------------------
		    find printers 
		*/
		{
			int j;
			while (!(mode & CNNL_JOB_CANCELLED)){
				
				memset(ipaddr, 0x00, STRING_SHORT);
				if (CNNL_GetIPAddress(CACHE_PATH, macaddr, ipaddr, STRING_SHORT, CNNET_SEARCH_AUTO, 1, 5000) == CNNL_RET_SUCCESS){
					break;
				}
				
				fprintf(stderr, "INFO: printer not responding; will retry in %d seconds...\n", WAITTIME_FIND_PRINTERS);
				for (j=0; j<WAITTIME_FIND_PRINTERS; j++, sleep(1)){
					// cancelled
					if ((mode & CNNL_JOB_CANCELLED)) goto error_init;
				}
				fprintf(stderr, "INFO: \n");
			}
			
			/* cancelled */
			if ((mode & CNNL_JOB_CANCELLED)) goto error_init;
		}
		fprintf(stderr, "INFO: \n");
		
		// wait a second
		sleep(1);
		
		/* ---------------------------------------------------------------------
		    open printer 
		*/
		if (CNNL_Init(&hnd) != CNNL_RET_SUCCESS){
			errcode = 1;
			fprintf(stderr, "ERROR: error occurred while preparing\n");
			goto error_init;
		}
		
		if (CNNL_Open(hnd, ipaddr) != CNNL_RET_SUCCESS){
			errcode = 1;
			fprintf(stderr, "ERROR: error occurred while preparing\n");
			goto error_open;
		}
		
		// check wether the printer exists or not
		if (CNNL_CheckVersion(hnd, 3, 5000) != CNNL_RET_SUCCESS){
			if (errcount==0){
				errcount++;
				CNNL_Close(hnd);
				CNNL_Terminate(&hnd);
				goto find_printer;
			} else {
				errcount++;
				fprintf(stderr, "ERROR: error occurred while preparing\n");
				errcode = 1;
				goto error_close;
			}
		} else {
			errcount=0;
		}

		/* ---------------------------------------------------------------------
		    start session
		*/
		mode = CNNL_JOB_WAITING;
		{
			int j=0, ret;
			
			if (!(mode & CNNL_JOB_CANCELLED)){ // if job is not cancelled
				while (!(mode & CNNL_JOB_CANCELLED)){
					// request to start session
					if ((ret=CNNL_SessionStart(hnd, user, hostname, job, 3, 3000)) == CNNL_RET_SUCCESS){
						// succeed to start session
						mode=CNNL_JOB_PRINTING;
						break;
					}
					
					// failed to start session
					if (ret == CNNL_RET_BUSY){
						// fprintf(stderr,"INFO: printer busy; will retry in %d seconds...\n", WAITTIME_START_SESSION);
						for (j=0; j<WAITTIME_START_SESSION; j++, sleep(1)){
							// cancelled 
							if ((mode & CNNL_JOB_CANCELLED)) goto error_close;
						}
					} else {
						fprintf(stderr, "INFO: printer not responding; will retry in %d seconds...\n", WAITTIME_RETRY_SEARCH);
						CNNL_Close(hnd);
						CNNL_Terminate(&hnd);
						for (j=0; j<WAITTIME_RETRY_SEARCH; j++, sleep(1)){
							// cancelled 
							if ((mode & CNNL_JOB_CANCELLED)) goto error_init;
						}
						goto find_printer;
					}
				}
			}
			if ((mode & CNNL_JOB_CANCELLED)){
				goto error_job_cancelled;
			}
		}
		fprintf(stderr, "INFO: \n");
		
		/* set timeout */
		CNNL_SetTimeout(hnd, SESSION_TIMEOUT, 3, 1000);
		
		/* start printing */
		if (!(mode & CNNL_JOB_ERROR) && CNNL_StartPrint(hnd, 3, 3000) != CNNL_RET_SUCCESS){
			errcode = 1;
			mode |= CNNL_JOB_ERROR;
			fprintf(stderr, "ERROR: error occurred while preparing\n");
			goto error_printing;
		}
		
		/* ---------------------------------------------------------------------
		    print(after this line, response error is fatal; job is cancelled)
		*/
		{
			// uint32_t d_time;
			// struct pollfd fds;
			int working=0, count=0, tmp_ret=CNNL_RET_SUCCESS;
			
			len = fread_with_retry(rbuffer, bufsize, fp);
			if ((mode & CNNL_JOB_ERROR)){
				errcode = 1;
				fprintf(stderr, "ERROR: error occurred while printing\n");
				goto error_close;
			}
			
			p = rbuffer;
			while (!(mode & CNNL_JOB_ERROR) && !(mode & CNNL_JOB_CANCELLED) && len > 0){
				
				ret = CNNL_DataWrite(hnd, p, len, &wlen, 6, 5000);
				written_size += wlen;
				
				if ((mode & CNNL_JOB_CANCELLED)) goto error_job_cancelled;
				
				if (ret == CNNL_RET_SUCCESS){
					len = fread_with_retry(rbuffer, bufsize, fp);
					p = rbuffer;
				} else if (ret == CNNL_RET_FAILURE){
					/* printing is failed */
					errcode = 1;
					mode |= CNNL_JOB_ERROR;
					fprintf(stderr, "ERROR: error occurred while printing\n");
					goto error_printing;
				} else {
					len -= wlen;
					p = p+wlen;
					wlen = 0;
				}

				/* check if printer is working */
				tmp_ret = CNNL_CheckPrint(hnd, 3, 1500);
				if (tmp_ret == CNNL_RET_SUCCESS){
					working = 1;
				} else if (tmp_ret == CNNL_RET_POWEROFF){
					goto error_job_cancelled;
				}
			}
			if ((mode & CNNL_JOB_CANCELLED)) goto error_job_cancelled;
			

			/* check if printer is working */
			for (count=0; working!=1 && count<10; sleep(1)){
				tmp_ret = CNNL_CheckPrint(hnd, 3, 3000);
				if (tmp_ret == CNNL_RET_SUCCESS){
					break;
				} else if (tmp_ret == CNNL_RET_FAILURE){
					goto error_printing;
				} else if (tmp_ret == CNNL_RET_POWEROFF){
					goto error_job_cancelled;
				} else {
					count++;
				}
				if ((mode & CNNL_JOB_CANCELLED)) goto error_job_cancelled;
			}

			/* finish to print */
			if (!(mode & CNNL_JOB_ERROR) && CNNL_EndPrint(hnd, 3, 3000) != CNNL_RET_SUCCESS){
				// cannot get end of print
				errcode = 1;
				mode |= CNNL_JOB_ERROR;
				fprintf(stderr, "ERROR: error occurred while printing\n");
				goto error_printing;
			}
		}
		

		/* ---------------------------------------------------------------------
		    close printer
		*/
error_job_cancelled:
		
error_printing:
		
		if (mode & CNNL_JOB_ERROR) {
			CNNL_SoftReset(hnd, 1, 3000);
		} else {
			if (mode & CNNL_JOB_CANCELLED){
				CNNL_SoftReset(hnd, 3, 3000);
			}
		}
		CNNL_SessionEnd(hnd, 3, 3000);
		
error_close:
		CNNL_Close(hnd);

error_open:
		CNNL_Terminate(&hnd);

error_init:
		fprintf(stderr, "INFO: \n");
		free(rbuffer);
	}
	
	// read buffer(flush)
	{
		unsigned char dummy;
		while (fread(&dummy, sizeof(unsigned char), 1, fp) > 0);
	}
	fflush(stdout);
	fflush(stderr);
	
	return (errcode == 0)? 0: 1;
}

static int _ij_network_search(void){
	CNNLHANDLE hnd=NULL, hmdl=NULL;
	int i=0, j=0, max = MAX_COUNT_OF_PRINTERS, found=0;
	CNNLNICINFO *nic;
	char model[STRING_SHORT],model2[STRING_SHORT], ipaddr[STRING_SHORT];
	
	if (CNNL_Init(&hmdl) != CNNL_RET_SUCCESS) goto error;

	// find printers
	nic = (CNNLNICINFO *)malloc(sizeof(CNNLNICINFO)*max);
	memset(nic, 0x00, sizeof(CNNLNICINFO)*max);
	
	if (CNNL_SearchPrinters(nic, CACHE_PATH, max, &found, CNNET_SEARCH_BROADCAST, 1, 5000) == CNNL_RET_SUCCESS){
		for (j=0; j<found; j++){
			
			memset(model, 0x00, sizeof(model));
			memset(model2, 0x00, sizeof(model2));
			
			memset(ipaddr, 0x00, STRING_SHORT);
			snprintf(ipaddr, STRING_SHORT-1, "%d.%d.%d.%d", 
					nic[j].ipaddr[0],nic[j].ipaddr[1],nic[j].ipaddr[2],nic[j].ipaddr[3]);
			
			if (CNNL_Open(hmdl, ipaddr) == CNNL_RET_SUCCESS){
				if (CNNL_GetModelName(hmdl, model, STRING_SHORT-1, 3, 1000) == CNNL_RET_SUCCESS){
					for (i=0; model[i]!='\0'; i++){
						if (model[i] == ' '){
							model2[i]='-';
						} else {
							model2[i]=model[i];
						}
					}
					
					printf("network cnijnet:/%02X-%02X-%02X-%02X-%02X-%02X \"%s\" \"%s_%02X-%02X-%02X-%02X-%02X-%02X\"\n", 
						nic[j].macaddr[0],nic[j].macaddr[1],nic[j].macaddr[2],
						nic[j].macaddr[3],nic[j].macaddr[4],nic[j].macaddr[5],
						model,
						model2,
						nic[j].macaddr[0],nic[j].macaddr[1],nic[j].macaddr[2],
						nic[j].macaddr[3],nic[j].macaddr[4],nic[j].macaddr[5]);
				}
				CNNL_Close(hmdl);
			}
		}
	} else {
		free(nic);
		goto error;
	}
	
	free(nic);
	CNNL_Terminate(&hmdl);
	
	return 0;
	
error:
	CNNL_Close(hnd);
	if (hmdl!= NULL) CNNL_Terminate(&hmdl);
	if (hnd != NULL) CNNL_Terminate(&hnd);

	return 0;
}

static void _ij_network_terminate(int sig_code){
	CNNL_Abort(hnd);
	mode |= CNNL_JOB_CANCELLED;
}

